# coding=utf-8

from typing import List, Union, NoReturn, Callable, Any, Text

from .acceleration import PocoAccelerationMixin
from .proxy import UIObjectProxy
from .agent import PocoAgent
from .gesture import PendingGestureAction
from .utils.track import MotionTrack


class Poco(PocoAccelerationMixin):
    def __init__(self, agent: PocoAgent, **options) -> Poco:
        self._agent = ...                           # type: PocoAgent
        self._pre_action_wait_for_appearance = 6
        self._post_action_interval = 0.8
        self._poll_interval = 1.44
        self._reevaluate_volatile_attributes = False # type: bool
        self._pre_action_callbacks = []             # type: List[Callable[Text, UIObjectProxy, Any]]
        self._post_action_callbacks = []            # type: List[Callable[Text, UIObjectProxy, Any]]

    def __call__(self, name: Text=None, **kw) -> UIObjectProxy:
        ...

    def wait_for_any(self, objects: List[UIObjectProxy], timeout: float=120.0) -> Union[NoReturn, UIObjectProxy]:
        ...

    def wait_for_all(self, objects: List[UIObjectProxy], timeout: float=120.0) -> NoReturn:
        ...

    def freeze(self) -> Poco:
        ...

    @property
    def agent(self) -> PocoAgent:
        ...

    def click(self, pos: (float, float)):
        ...

    def swipe(self, p1: (float, float), p2: (float, float)=None, direction: (float, float)=None, duration: float=2.0):
        ...

    def long_click(self, pos: (float, float), duration: float=2.0):
        ...

    def scroll(self, direction: Union[List[float], str], percent: float=0.6, duration: float=2.0):
        ...

    def pinch(self, direction: str='in', percent: float=0.6, duration: float=2.0, dead_zone: float=0.1):
        ...

    def pan(self, direction: (float, float), duration: float=2.0):
        ...

    def start_gesture(self, pos: (float, float)) -> PendingGestureAction:
        ...

    def apply_motion_tracks(self, tracks: List[MotionTrack], accuracy: float=0.004):
        ...

    def snapshot(self, width: int=720) -> (str, str):
        ...

    def get_screen_size(self) -> (float, float):
        ...

    def wait_stable(self):
        ...

    def sleep_for_polling_interval(self):
        ...

    def on_pre_action(self, action: Text, ui: UIObjectProxy, args: Any) -> NoReturn:
        ...
    def on_post_action(self, action: Text, ui: UIObjectProxy, args: Any) -> NoReturn:
        ...
    def pre_action(self, action: Text, ui: UIObjectProxy, args: Any) -> NoReturn:
        ...
    def post_action(self, ction: Text, ui: UIObjectProxy, args: Any) -> NoReturn:
        ...



# coding=utf-8


from poco.sdk.AbstractDumper import AbstractDumper
from poco.drivers.osx.sdk.OSXUINode import OSXUINode
from poco.sdk.exceptions import InvalidSurfaceException


class OSXUIDumper(AbstractDumper):

    def __init__(self, root):
        self.RootControl = root
        self.RootHeight = self.RootControl.AXSize[1]
        self.RootWidth = self.RootControl.AXSize[0]
        self.RootLeft = self.RootControl.AXPosition[0]
        self.RootTop = self.RootControl.AXPosition[1]
        if self.RootWidth == 0 or self.RootHeight == 0:
            raise InvalidSurfaceException(self, "You may have minimized your window or the window is too small!")

    def getRoot(self):
        return OSXUINode(self.RootControl, self)



# coding=utf-8

import time
import base64
import zlib
import re
import pyautogui
import atomac
import operator
from pynput.keyboard import Controller
from poco.sdk.std.rpc.controller import StdRpcEndpointController
from poco.sdk.std.rpc.reactor import StdRpcReactor
from poco.utils.net.transport.tcp import TcpSocket
from poco.drivers.osx.sdk.OSXUIDumper import OSXUIDumper
from poco.sdk.exceptions import UnableToSetAttributeException, NonuniqueSurfaceException, InvalidSurfaceException
from poco.utils.six import string_types, PY2
from poco.utils.six.moves import reduce
from poco.drivers.osx.sdk.OSXUIFunc import OSXFunc

DEFAULT_PORT = 15004
DEFAULT_ADDR = ('0.0.0.0', DEFAULT_PORT)


class PocoSDKOSX(object):

    def __init__(self, addr=DEFAULT_ADDR):
        self.reactor = None
        self.addr = addr
        self.running = False
        self.root = None
        self.keyboard = Controller()

    def Dump(self, _):
        res = OSXUIDumper(self.root).dumpHierarchy()
        return res

    def SetForeground(self):
        self.root.AXMinimized = False
        self.app.AXFrontmost = True

    def GetSDKVersion(self):
        return '0.0.1'

    def GetDebugProfilingData(self):
        return {}

    def GetScreenSize(self):
        Width = self.root.AXSize[0]
        Height = self.root.AXSize[1]
        return [Width, Height]

    def GetWindowRect(self):
        Width = self.root.AXSize[0]
        Height = self.root.AXSize[1]
        return [self.root.AXPosition[0], self.root.AXPosition[1], self.root.AXPosition[0] + Width, self.root.AXPosition[1] + Height]

    def KeyEvent(self, keycode):
        waittime = 0.05
        for c in keycode:
            self.keyboard.press(key=c)
            self.keyboard.release(key=c)
            time.sleep(waittime)
        return True

    def Screenshot(self, width):
        self.SetForeground()
        size = self.root.AXSize
        pos = self.root.AXPosition
        pyautogui.screenshot('Screenshot.png', (pos[0], pos[1], size[0], size[1])).save('Screenshot.png')
        f = open(r'Screenshot.png', 'rb')
        deflated = zlib.compress(f.read())
        ls_f = base64.b64encode(deflated)
        f.close()
        return [ls_f, "png.deflate"]

        # self.root.ToBitmap().ToFile('Screenshot.bmp')
        # f = open(r'Screenshot.bmp', 'rb')
        # ls_f = base64.b64encode(f.read())
        # f.close()
        # return [ls_f, "bmp"]

    def Click(self, x, y):
        self.SetForeground()
        size = self.root.AXSize
        pos = self.root.AXPosition
        OSXFunc.click(pos[0] + size[0] * x, pos[1] + size[1] * y)
        return True

    def RClick(self, x, y):
        self.SetForeground()
        size = self.root.AXSize
        pos = self.root.AXPosition
        OSXFunc.rclick(pos[0] + size[0] * x, pos[1] + size[1] * y)
        return True

    def DoubleClick(self, x, y):
        self.SetForeground()
        size = self.root.AXSize
        pos = self.root.AXPosition
        OSXFunc.doubleclick(pos[0] + size[0] * x, pos[1] + size[1] * y)   
        return True

    def Swipe(self, x1, y1, x2, y2, duration):
        self.SetForeground()
        Left = self.root.AXPosition[0]
        Top = self.root.AXPosition[1]
        Width = self.root.AXSize[0]
        Height = self.root.AXSize[1]
        x1 = Left + Width * x1
        y1 = Top + Height * y1
        x2 = Left + Width * x2
        y2 = Top + Height * y2
        sx = abs(x1 - x2)
        sy = abs(y1 - y2)
        stepx = sx / (duration * 10.0)  # 将滑动距离分割，实现平滑的拖动
        stepy = sy / (duration * 10.0)
        OSXFunc.move(x1, y1)
        OSXFunc.press(x1, y1)
        duration = int(duration * 10.0)
        for i in range(duration + 1):
            OSXFunc.drag(x1 + stepx * i, y1 + stepy * i)
            time.sleep(0.1)
        OSXFunc.release(x2, y2)
        return True

    def LongClick(self, x, y, duration, **kwargs):
        self.SetForeground()

        # poco std暂不支持选择鼠标按键
        button = kwargs.get("button", "left")
        if button not in ("left", "right"):
            raise ValueError("Unknow button: " + button)
        if button is "left":
            button = 1
        else:
            button = 2

        Left = self.root.AXPosition[0]
        Top = self.root.AXPosition[1]
        Width = self.root.AXSize[0]
        Height = self.root.AXSize[1]
        x = Left + Width * x
        y = Top + Height * y
        OSXFunc.move(x, y)
        OSXFunc.press(x, y, button=button)
        time.sleep(duration)
        OSXFunc.release(x, y, button=button)
        return True

    def Scroll(self, direction, percent, duration):
        if direction not in ('vertical', 'horizontal'):
            raise ValueError('Argument `direction` should be one of "vertical" or "horizontal". Got {}'.format(repr(direction)))

        x = 0.5  # 先把鼠标移到窗口中间，这样才能保证滚动的是这个窗口。
        y = 0.5
        steps = percent
        Left = self.GetWindowRect()[0]
        Top = self.GetWindowRect()[1]
        Width = self.GetScreenSize()[0]
        Height = self.GetScreenSize()[1]
        x = Left + Width * x
        y = Top + Height * y
        x = int(x)
        y = int(y)
        OSXFunc.move(x, y)

        if direction == 'horizontal':
            interval = float(duration) / (abs(steps) + 1)
            if steps < 0:
                for i in range(0, abs(steps)):
                    time.sleep(interval)
                    OSXFunc.scroll(None, 1)
            else:
                for i in range(0, abs(steps)):
                    time.sleep(interval)
                    OSXFunc.scroll(None, -1)
        else:
            interval = float(duration) / (abs(steps) + 1)
            if steps < 0:
                for i in range(0, abs(steps)):
                    time.sleep(interval)
                    OSXFunc.scroll(1)
            else:
                for i in range(0, abs(steps)):
                    time.sleep(interval)
                    OSXFunc.scroll(-1)
        return True

    def EnumWindows(self, selector):
        names = []  # 一个应用程序会有多个窗口，因此我们要先枚举一个应用程序里的所有窗口
        if 'bundleid' in selector:
            self.app = OSXFunc.getAppRefByBundleId(selector['bundleid'])
            windows = self.app.windows()
            for i, w in enumerate(windows):
                names.append((w.AXTitle, i))
            return names

        if 'appname' in selector:
            self.app = OSXFunc.getAppRefByLocalizedName(selector['appname'])
            windows = self.app.windows()
            for i, w in enumerate(windows):
                names.append((w.AXTitle, i))
            return names

        if 'appname_re' in selector:  # 此方法由于MacOS API，问题较多
            apps = OSXFunc.getRunningApps()  # 获取当前运行的所有应用程序
            appset = []  # 应用程序集合
            appnameset = set()  # 应用程序标题集合
            for t in apps:
                tempapp = OSXFunc.getAppRefByPid(t.processIdentifier())
                if str(tempapp) == str(atomac.AXClasses.NativeUIElement()):  # 通过trick判断应用程序是都否为空
                    continue
                attrs = tempapp.getAttributes()
                if 'AXTitle' in attrs:
                    tit = tempapp.AXTitle
                    if re.match(selector['appname_re'], tit):
                        appset.append(tempapp)
                        appnameset.add(tit)  # 这里有Bug，可能会获取到进程的不同副本，所以要通过名字去判断是否唯一

            if len(appnameset) is 0:
                raise InvalidSurfaceException(selector, "Can't find any applications by the given parameter")
            if len(appnameset) != 1:
                raise NonuniqueSurfaceException(selector)
            while len(names) is 0:  # 有可能有多个副本，但只有一个真的应用程序有窗口，所以要枚举去找
                if len(appset) is 0:
                    return names
                self.app = appset.pop()
                windows = self.app.windows()  # 获取当前应用程序的所有窗口
                for i, w in enumerate(windows):
                    names.append((w.AXTitle, i))
            return names
        return names
        
    def ConnectWindowsByWindowTitle(self, selector, wlist):
        hn = set()
        for n in wlist:
            if selector['windowtitle'] == n[0]:
                hn.add(n[1])  # 添加窗口索引到集合里
        if len(hn) == 0:
            return -1
        return hn

    def ConnectWindowsByWindowTitleRe(self, selector, wlist):
        hn = set()
        for n in wlist:
            if re.match(selector['windowtitle_re'], n[0]):
                hn.add(n[1])  # 添加窗口索引到集合里
        if len(hn) == 0:
            return -1
        return hn

    def ConnectWindow(self, selector):
        
        # 目前来说，如下处理，以后添加更多的参数后需修改代码逻辑
        argunums = 0
        if 'bundleid' in selector:
            argunums += 1
        if 'appname' in selector:
            argunums += 1
        if 'appname_re' in selector:
            argunums += 1
        if argunums == 0:
            raise ValueError("Expect bundleid or appname, got none")
        elif argunums != 1:
            raise ValueError("Too many arguments, only need bundleid or appname or appname_re")
        
        winlist = self.EnumWindows(selector)

        handleSetList = []
        if 'windowtitle' in selector:
            handleSetList.append(self.ConnectWindowsByWindowTitle(selector, winlist))
        if 'windowindex' in selector:
            handleSetList.append(set([selector['windowindex']]))
        if "windowtitle_re" in selector:
            handleSetList.append(self.ConnectWindowsByWindowTitleRe(selector, winlist))

        while -1 in handleSetList:
            handleSetList.remove(-1)  # 有些参数没有提供会返回-1.把所有的-1去掉

        if len(handleSetList) == 0:  # 三种方法都找不到窗口
            raise InvalidSurfaceException(selector, "Can't find any applications by the given parameter")
            
        handleSet = reduce(operator.__and__, handleSetList)  # 提供了多个参数来确定唯一一个窗口，所以要做交集，取得唯一匹配的窗口

        if len(handleSet) == 0:
            raise InvalidSurfaceException(selector, "Can't find any applications by the given parameter")
        elif len(handleSet) != 1:
            raise NonuniqueSurfaceException(selector)
        else:
            hn = handleSet.pop()  # 取得该窗口的索引
            w = self.app.windows()
            if len(w) <= hn:
                raise IndexError("Unable to find the specified window through the index, you may have closed the specified window during the run")
            self.root = self.app.windows()[hn]
            self.SetForeground()
            return True

    def run(self):
        self.reactor = StdRpcReactor()
        self.reactor.register('Dump', self.Dump)
        self.reactor.register('GetSDKVersion', self.GetSDKVersion)
        self.reactor.register('GetDebugProfilingData', self.GetDebugProfilingData)
        self.reactor.register('GetScreenSize', self.GetScreenSize)
        self.reactor.register('Screenshot', self.Screenshot)
        self.reactor.register('Click', self.Click)
        self.reactor.register('Swipe', self.Swipe)
        self.reactor.register('LongClick', self.LongClick)
        self.reactor.register('SetForeground', self.SetForeground)
        self.reactor.register('ConnectWindow', self.ConnectWindow)
        self.reactor.register('Scroll', self.Scroll)
        self.reactor.register('RClick', self.RClick)
        self.reactor.register('DoubleClick', self.DoubleClick)
        self.reactor.register('KeyEvent', self.KeyEvent)
        transport = TcpSocket()
        transport.bind(self.addr)
        self.rpc = StdRpcEndpointController(transport, self.reactor)
        if self.running is False:
            self.running = True
            self.rpc.serve_forever()

if __name__ == '__main__':
    pocosdk = PocoSDKOSX()
    # pocosdk.ConnectWindow({'appname_re': u'系统偏好', 'windowindex': 0})
    # pocosdk.ConnectWindow({'appname': u'系统偏好设置', 'windowtitle': u'系统偏好设置'})
    # pocosdk.ConnectWindow({'bundleid': u'com.microsoft.VSCode', 'windowindex': 0})
    
    pocosdk.run()



# coding=utf-8
import websocket
from threading import Thread
from ..interfaces import IClient
from poco.utils import six


DEFAULT_ADDR = "ws://localhost:5003"


class WebSocketClient(IClient):

    def __init__(self, addr=DEFAULT_ADDR):
        super(WebSocketClient, self).__init__()
        self.addr = addr
        self._inbox = []
        self._ws = None
        self._ws_thread = None

    def __str__(self):
        return self.addr
    __repr__ = __str__

    def connect(self):
        if self._ws_thread:
            self.close()
        print("connecting server..")
        self._init_ws_thread()

    def send(self, msg):
        if not isinstance(msg, six.text_type):
            msg = msg.decode("utf-8")
        self._ws.send(msg)

    def recv(self):
        msgs, self._inbox = self._inbox, []
        return msgs

    def close(self):
        print("closing connection..")
        self._ws.close()
        self._ws_thread = None

    def _init_ws_thread(self):
        self._ws = self._init_ws()
        t = Thread(target=self._ws.run_forever)
        t.daemon = True
        t.start()
        self._ws_thread = t

    def _init_ws(self):
        ws = websocket.WebSocketApp(self.addr,
                                    on_open=self._on_ws_open,
                                    on_message=self._on_ws_message,
                                    on_error=self._on_ws_error,
                                    on_close=self._on_ws_close)
        # ws.enableTrace(True)
        return ws

    def _on_ws_message(self, ws, message):
        self._inbox.append(message)

    def _on_ws_error(self, ws, error):
        print("on error", error)
        self.on_close()

    def _on_ws_close(self, ws, *args, **kwargs):
        print("on close")
        self.on_close()

    def _on_ws_open(self, ws):
        print('on open')
        self.on_connect()



# coding=utf-8

import time
from poco.drivers.unity3d.test.tutorial.case import TutorialCase


class LocalPositioning1Tutorial(TutorialCase):
    def runTest(self):
        image = self.poco('fish').child(type='Image')
        image.focus('center').long_click()
        time.sleep(0.2)
        image.focus([0.1, 0.1]).long_click()
        time.sleep(0.2)
        image.focus([0.9, 0.9]).long_click()
        time.sleep(0.2)
        image.focus([0.5, 0.9]).long_click()
        time.sleep(0.2)


if __name__ == '__main__':
    import pocounit
    pocounit.main()



# coding=utf-8

from poco.utils.airtest import AirtestInput
from poco.sdk.interfaces.hierarchy import HierarchyInterface
from poco.sdk.interfaces.input import InputInterface
from poco.sdk.interfaces.screen import ScreenInterface
from poco.sdk.interfaces.command import CommandInterface

__author__ = 'lxn3032'


def _assign(val, default_val):
    if isinstance(val, type(None)):
        return default_val
    else:
        return val


class PocoAgent(object):
    """
    This is the agent class for poco to communicate with target device.

    This class is an aggregation of 4 major interfaces for now.

    - :py:class:`HierarchyInterface <poco.sdk.interfaces.hierarchy.HierarchyInterface>`: defines the hierarchy
      accessibility methods such as dump(crawl the whole UI tree), getAttr(retrieve attribute value by name)
    - :py:class:`InputInterface <poco.sdk.interfaces.input.InputInterface>`: defines the simulated input methods to
      allow inject simulated input on target device
    - :py:class:`ScreenInterface <poco.sdk.interfaces.screen.ScreenInterface>`: defines methods to access the screen
      surface
    - :py:class:`CommandInterface <poco.sdk.interfaces.command.CommandInterface>`: defines methods to communicate
      with target device in arbitrary way. This is optional.
    """

    def __init__(self, hierarchy, input, screen, command=None):
        self.hierarchy = _assign(hierarchy, HierarchyInterface())
        self.input = _assign(input, InputInterface())
        self.screen = _assign(screen, ScreenInterface())
        self.command = _assign(command, CommandInterface())
        self._driver = None

    def get_sdk_version(self):
        """
        Retrieve the sdk version from remote runtime. Each poco agent implementation should override this method.

        Returns:
            :py:obj:`str`: version string of the poco sdk. usually in "0.0.0" format. None if not provided by poco sdk.
        """
        pass

    def rpc_reconnect(self):
        self.rpc.close()
        self.rpc.connect()

    @property
    def rpc(self):
        """
        Return the interface of this agent handled.

        Returns:
            :py:obj:`object`: the rpc interface of this agent handled.

        Raises:
            NotImplementedError: raises if the agent implementation dose not expose the rpc interface to user.
        """

        raise NotImplementedError('This poco agent does not have a explicit rpc connection.')

    def on_bind_driver(self, driver):
        self._driver = driver
        if isinstance(self.input, AirtestInput):
            self.input.add_preaction_cb(driver)

    @property
    def driver(self):
        """
        Return the driver of this agent related to. None if the driver is not ready to bind.

        Returns:
            :py:class:`inherit from Poco <poco.pocofw.Poco>`: the driver of this agent related to.
        """

        if not self._driver:
            raise AttributeError("`driver` is not bound on this agent implementation({}). "
                                 "Do you forget to call `super().on_bind_driver` when you override the method "
                                 "`on_bind_driver` in your sub class?"
                                 .format(repr(self)))
        return self._driver



# -*- coding: utf-8 -*-
# @Author: gzliuxin
# @Email:  gzliuxin@corp.netease.com
# @Date:   2017-07-14 19:47:51

from poco.drivers.std import StdPoco
from poco.drivers.unity3d.device import UnityEditorWindow
from poco.exceptions import InvalidOperationException
from airtest.core.api import connect_device, device as current_device

__all__ = ['UnityPoco']
DEFAULT_PORT = 5001
DEFAULT_ADDR = ("localhost", DEFAULT_PORT)


class UnityVRSupport():
    def __init__(self, client):
        self.client = client
        self.support_vr = False
        try:
            self.support_vr = self.client.call("isVrSupported")
        except InvalidOperationException:
            raise InvalidOperationException('VR not supported')

    def hasMovementFinished(self):
        success, error_msg  =self.client.call("hasMovementFinished").wait()
        print(success)
        if success != None:
            return True
        else:
            return False

    def rotateObject(self, x, y, z, camera, follower, speed=0.125):
        return self.client.call("RotateObject", x, y, z, camera, follower, speed)

    def objectLookAt(self, name, camera, follower, speed=0.125):
        return self.client.call("ObjectLookAt", name, camera, follower, speed)


class UnityPoco(StdPoco):
    """
    Poco Unity3D implementation.

    Args:
        addr (:py:obj:`tuple`): the endpoint of your Unity3D game, default to ``("localhost", 5001)``
        unity_editor (:py:obj:`bool`): whether your Unity3D game is running in UnityEditor or not. default to ``False``
        connect_default_device (:py:obj:`bool`): whether connect to a default device if no devices selected manually.
         default to ``True``.
         device (:py:obj:`Device`): :py:obj:`airtest.core.device.Device` instance provided by ``airtest``. leave the
         parameter default and the default device will be chosen. more details refer to ``airtest doc``
        options: see :py:class:`poco.pocofw.Poco`

    Examples:
        If your game is running on Android, you could initialize poco instance by using following snippet::
            
            from poco.drivers.unity3d import UnityPoco
            
            # your phone and your PC/mac should be inside the same sub-net.
            ip = '<ip address of your phone>'
            poco = UnityPoco((ip, 5001))
            poco('button').click()
            ...

    """

    def __init__(self, addr=DEFAULT_ADDR, unity_editor=False, connect_default_device=True, device=None, **options):
        if 'action_interval' not in options:
            options['action_interval'] = 0.5

        if unity_editor:
            dev = UnityEditorWindow()
        else:
            dev = device or current_device()

        if dev is None and connect_default_device and not current_device():
            # currently only connect to Android as default
            # can apply auto detection in the future
            dev = connect_device("Android:///")

        super(UnityPoco, self).__init__(addr[1], dev, ip=addr[0], **options)
        # If some devices fail to initialize, the UI tree cannot be obtained
        # self.vr = UnityVRSupport(self.agent.rpc)

    def send_message(self, message):
        self.agent.rpc.call("SendMessage", message)

    def invoke(self, listener, **kwargs):
        callback = self.agent.rpc.call("Invoke", listener=listener, data=kwargs)

        value, error = callback.wait()

        if error is not None:
            raise Exception(error)

        return value



# coding=utf-8
import xml.etree.ElementTree as ET
from poco.pocofw import Poco
from poco.agent import PocoAgent
from poco.freezeui.hierarchy import FrozenUIDumper, FrozenUIHierarchy
from poco.utils.airtest import AirtestInput, AirtestScreen
from poco.utils import six
from airtest.core.api import device as current_device
try:
    from airtest.core.ios.rotation import XYTransformer
except AttributeError:
    raise RuntimeError('The iOS module of Airtest>1.1.7 only supports python3, if you want to use, please upgrade to python3 first.')
from pprint import pprint


class iosPoco(Poco):
    def __init__(self, device=None, **kwargs):
        device = device or current_device()
        if not device or device.__class__.__name__ != 'IOS':
            raise RuntimeError('Please call `connect_device` to connect an iOS device first')

        agent = iosPocoAgent(device)
        super(iosPoco, self).__init__(agent, **kwargs)


class iosPocoAgent(PocoAgent):
    def __init__(self, client):
        self.client = client

        hierarchy = FrozenUIHierarchy(iosDumper(self.client))
        screen = AirtestScreen()
        input = AirtestInput()
        super(iosPocoAgent, self).__init__(hierarchy, input, screen, None)


class iosDumper(FrozenUIDumper):

    def __init__(self, client):
        super(iosDumper, self).__init__()
        self.client = client
        self.size = client.display_info["window_width"], client.display_info["window_height"]

    def dumpHierarchy(self, onlyVisibleNode=True):
        # 当使用了appium/WebDriverAgent时，ipad横屏且在桌面下，坐标需要按照竖屏坐标额外转换一次
        # 判断条件如下：
        # 当ios.using_ios_tagent有值、且为false，说明使用的是appium/WebDriverAgent
        # airtest<=1.2.4时，没有ios.using_ios_tagent的值，说明也是用的appium/wda
        if ((hasattr(self.client, "using_ios_tagent") and not self.client.using_ios_tagent) or
            (not hasattr(self.client, "using_ios_tagent"))) \
                and (self.client.is_pad and self.client.orientation != 'PORTRAIT' and self.client.home_interface()):
            switch_flag = True
        else:
            switch_flag = False
        jsonObj = self.client.driver.source(format='json')
        w, h = self.size
        if self.client.orientation in ['LANDSCAPE', 'UIA_DEVICE_ORIENTATION_LANDSCAPERIGHT']:
            w, h = h, w
        data = json_parser(jsonObj, (w, h), switch_flag=switch_flag, ori=self.client.orientation)
        return data

    def dumpHierarchy_xml(self):
        xml = self.client.driver.source()
        xml = xml.encode("utf-8")
        data = ios_dump_xml(xml, self.size)
        return data


def ios_dump_xml(xml, screen_size):
    root = ET.fromstring(xml)
    data = xml_parser(root, screen_size)
    return data


def json_parser(node, screen_size, switch_flag=False, ori='PORTRAIT'):
    """

    :param node: node info {}
    :param screen_size: ios.windows_size()
    :param switch_flag: If it is an ipad , when on the desktop, all coordinates must be converted to vertical screen coordinates
    :param ori: wda ['PORTRAIT', 'LANDSCAPE',
                    'UIA_DEVICE_ORIENTATION_LANDSCAPERIGHT',
                    'UIA_DEVICE_ORIENTATION_PORTRAIT_UPSIDEDOWN']
    :return:
    """
    screen_w, screen_h = screen_size

    if "name" in node and node["name"]:
        name = node["name"]
    else:
        name = node["type"]

    data = {
        "name": name,
        "payload": {}
    }

    if six.PY2:
        for key in [x for x in node.keys() if x not in ['frame', 'children']]:
            data["payload"][key.encode("utf-8")] = node[key]
    else:
        for key in [x for x in node.keys() if x not in ['frame', 'children']]:
            data["payload"][key] = node[key]

    w = float(node["rect"]["width"])
    h = float(node["rect"]["height"])
    x = float(node["rect"]["x"])
    y = float(node["rect"]["y"])

    if switch_flag:
        x, y = XYTransformer.ori_2_up(
            (x, y),
            (screen_w, screen_h),
            ori
        )
    if switch_flag and ori == 'LANDSCAPE':
        w, h = h, w
        data["payload"]["pos"] = [
            (x + w / 2) / screen_w,
            (y - h / 2) / screen_h
        ]
    elif switch_flag and ori == 'UIA_DEVICE_ORIENTATION_LANDSCAPERIGHT':
        w, h = h, w
        data["payload"]["pos"] = [
            (x - w / 2) / screen_w,
            (y + h / 2) / screen_h
        ]
    elif switch_flag and ori == 'UIA_DEVICE_ORIENTATION_PORTRAIT_UPSIDEDOWN':
        data["payload"]["pos"] = [
            (x - w / 2) / screen_w,
            (y - h / 2) / screen_h
        ]
    else:
        data["payload"]["pos"] = [
            (x + w / 2) / screen_w,
            (y + h / 2) / screen_h
        ]

    data["payload"]["name"] = name
    data["payload"]["size"] = [w / screen_w, h / screen_h]
    
    data["payload"]["zOrders"] = {
        "local": 0,
        "global": 0,
    }
    data["payload"]["anchorPoint"] = [0.5, 0.5]

    # TODO: w = 0 and h = 0 situation need to solve with
    # roll back set as True when finding a visible child
    if "visible" not in node:
        if (x >= 0 or x + w > 0) and (x < screen_w) and (y >= 0 or y + h > 0) and (y < screen_h):
            data["payload"]["visible"] = True
        elif w == 0 or h == 0:
            data["payload"]["visible"] = True
        else:
            data["payload"]["visible"] = False

    children_data = []
    if "children" in node:
        for child in node["children"]:
            child_data = json_parser(child, screen_size=screen_size, switch_flag=switch_flag, ori=ori)
            children_data.append(child_data)

    if children_data:
        data["children"] = children_data
        if data["payload"]["visible"] is False:
            for child_node in children_data:
                if child_node["payload"].get("visible") is True:
                    data["payload"]["visible"] = True
                    break

    return data


def xml_parser(node, screen_size):
    # print(node)
    # print(node.attrib)
    screen_w, screen_h = screen_size

    name = node.attrib.get("name", node.tag)
    data = {
        "name": name,
        "payload": node.attrib,
    }

    w = float(node.attrib["width"])
    h = float(node.attrib["height"])
    x = float(node.attrib["x"])
    y = float(node.attrib["y"])

    data["payload"]["name"] = name
    data["payload"]["size"] = [w / screen_w, h / screen_h]
    data["payload"]["pos"] = [
        (x + w / 2) / screen_w,
        (y + h / 2) / screen_h
    ]
    data["payload"]["zOrders"] = {
        "local": 0,
        "global": 0,
    }
    data["payload"]["anchorPoint"] = (0.5, 0.5)

    children_data = []
    for child in node:
        child_data = xml_parser(child, screen_size)
        children_data.append(child_data)

    if children_data:
        data["children"] = children_data

    return data


if __name__ == '__main__':
    XML = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<XCUIElementTypeApplication type=\"XCUIElementTypeApplication\" name=\"健康\" label=\"健康\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n  <XCUIElementTypeWindow type=\"XCUIElementTypeWindow\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n    <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n      <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n        <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n          <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n            <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n              <XCUIElementTypeNavigationBar type=\"XCUIElementTypeNavigationBar\" name=\"WDTodayDayPageView\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"44\" width=\"375\" height=\"44\">\n                <XCUIElementTypeButton type=\"XCUIElementTypeButton\" name=\"五月\" label=\"五月\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"44\" width=\"65\" height=\"44\"/>\n                <XCUIElementTypeButton type=\"XCUIElementTypeButton\" name=\"查看个人资料\" label=\"查看个人资料\" enabled=\"true\" visible=\"true\" x=\"334\" y=\"53\" width=\"25\" height=\"26\"/>\n              </XCUIElementTypeNavigationBar>\n              <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n                <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n                  <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n                    <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n                      <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n                        <XCUIElementTypeTable type=\"XCUIElementTypeTable\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n                          <XCUIElementTypeOther type=\"XCUIElementTypeOther\" name=\"健身记录\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"176\" width=\"375\" height=\"37\">\n                            <XCUIElementTypeOther type=\"XCUIElementTypeOther\" name=\"健身记录\" label=\"健身记录\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"176\" width=\"375\" height=\"37\"/>\n                          </XCUIElementTypeOther>\n                          <XCUIElementTypeCell type=\"XCUIElementTypeCell\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"213\" width=\"375\" height=\"82\">\n                            <XCUIElementTypeStaticText type=\"XCUIElementTypeStaticText\" value=\"1.8￼公里\" name=\"1.8￼公里\" label=\"1.8￼公里\" enabled=\"true\" visible=\"true\" x=\"254\" y=\"214\" width=\"93\" height=\"54\"/>\n                            <XCUIElementTypeStaticText type=\"XCUIElementTypeStaticText\" value=\"今天 上午10:05\" name=\"今天 上午10:05\" label=\"今天 上午10:05\" enabled=\"true\" visible=\"true\" x=\"262\" y=\"266\" width=\"86\" height=\"15\"/>\n                            <XCUIElementTypeStaticText type=\"XCUIElementTypeStaticText\" value=\"步行 + 跑步距离\" name=\"步行 + 跑步距离\" label=\"步行 + 跑步距离\" enabled=\"true\" visible=\"true\" x=\"28\" y=\"220\" width=\"124\" height=\"21\"/>\n                          </XCUIElementTypeCell>\n                          <XCUIElementTypeCell type=\"XCUIElementTypeCell\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"295\" width=\"375\" height=\"82\">\n                            <XCUIElementTypeStaticText type=\"XCUIElementTypeStaticText\" value=\"2,438￼步\" name=\"2,438￼步\" label=\"2,438￼步\" enabled=\"true\" visible=\"true\" x=\"209\" y=\"296\" width=\"138\" height=\"54\"/>\n                            <XCUIElementTypeStaticText type=\"XCUIElementTypeStaticText\" value=\"今天 上午10:05\" name=\"今天 上午10:05\" label=\"今天 上午10:05\" enabled=\"true\" visible=\"true\" x=\"262\" y=\"348\" width=\"86\" height=\"15\"/>\n                            <XCUIElementTypeStaticText type=\"XCUIElementTypeStaticText\" value=\"步数\" name=\"步数\" label=\"步数\" enabled=\"true\" visible=\"true\" x=\"28\" y=\"302\" width=\"35\" height=\"21\"/>\n                          </XCUIElementTypeCell>\n                          <XCUIElementTypeCell type=\"XCUIElementTypeCell\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"377\" width=\"375\" height=\"82\">\n                            <XCUIElementTypeStaticText type=\"XCUIElementTypeStaticText\" value=\"今天 上午9:24\" name=\"今天 上午9:24\" label=\"今天 上午9:24\" enabled=\"true\" visible=\"true\" x=\"268\" y=\"430\" width=\"79\" height=\"15\"/>\n                            <XCUIElementTypeStaticText type=\"XCUIElementTypeStaticText\" value=\"4￼层\" name=\"4￼层\" label=\"4￼层\" enabled=\"true\" visible=\"true\" x=\"300\" y=\"378\" width=\"48\" height=\"54\"/>\n                            <XCUIElementTypeStaticText type=\"XCUIElementTypeStaticText\" value=\"已爬楼层\" name=\"已爬楼层\" label=\"已爬楼层\" enabled=\"true\" visible=\"true\" x=\"28\" y=\"384\" width=\"70\" height=\"21\"/>\n                          </XCUIElementTypeCell>\n                        </XCUIElementTypeTable>\n                      </XCUIElementTypeOther>\n                    </XCUIElementTypeOther>\n                  </XCUIElementTypeOther>\n                </XCUIElementTypeOther>\n              </XCUIElementTypeOther>\n              <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n                <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"88\" width=\"375\" height=\"88\">\n                  <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"88\" width=\"375\" height=\"88\">\n                    <XCUIElementTypeScrollView type=\"XCUIElementTypeScrollView\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"105\" width=\"375\" height=\"35\">\n                      <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"105\" width=\"375\" height=\"58\">\n                        <XCUIElementTypeOther type=\"XCUIElementTypeOther\" name=\"5月13日 星期日\" label=\"5月13日 星期日\" enabled=\"true\" visible=\"true\" x=\"13\" y=\"110\" width=\"36\" height=\"35\"/>\n                        <XCUIElementTypeOther type=\"XCUIElementTypeOther\" name=\"今天, 5月14日 星期一\" label=\"今天, 5月14日 星期一\" enabled=\"true\" visible=\"true\" x=\"65\" y=\"110\" width=\"36\" height=\"35\"/>\n                        <XCUIElementTypeOther type=\"XCUIElementTypeOther\" name=\"5月15日 星期二\" label=\"5月15日 星期二\" enabled=\"true\" visible=\"true\" x=\"118\" y=\"110\" width=\"35\" height=\"35\"/>\n                        <XCUIElementTypeOther type=\"XCUIElementTypeOther\" name=\"5月16日 星期三\" label=\"5月16日 星期三\" enabled=\"true\" visible=\"true\" x=\"170\" y=\"110\" width=\"35\" height=\"35\"/>\n                        <XCUIElementTypeOther type=\"XCUIElementTypeOther\" name=\"5月17日 星期四\" label=\"5月17日 星期四\" enabled=\"true\" visible=\"true\" x=\"222\" y=\"110\" width=\"35\" height=\"35\"/>\n                        <XCUIElementTypeOther type=\"XCUIElementTypeOther\" name=\"5月18日 星期五\" label=\"5月18日 星期五\" enabled=\"true\" visible=\"true\" x=\"274\" y=\"110\" width=\"36\" height=\"35\"/>\n                        <XCUIElementTypeOther type=\"XCUIElementTypeOther\" name=\"5月19日 星期六\" label=\"5月19日 星期六\" enabled=\"true\" visible=\"true\" x=\"326\" y=\"110\" width=\"36\" height=\"35\"/>\n                      </XCUIElementTypeOther>\n                    </XCUIElementTypeScrollView>\n                    <XCUIElementTypeStaticText type=\"XCUIElementTypeStaticText\" value=\"2018年5月14日 星期一\" name=\"2018年5月14日 星期一\" label=\"2018年5月14日 星期一\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"148\" width=\"375\" height=\"20\"/>\n                  </XCUIElementTypeOther>\n                </XCUIElementTypeOther>\n              </XCUIElementTypeOther>\n            </XCUIElementTypeOther>\n          </XCUIElementTypeOther>\n        </XCUIElementTypeOther>\n      </XCUIElementTypeOther>\n      <XCUIElementTypeTabBar type=\"XCUIElementTypeTabBar\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"729\" width=\"375\" height=\"83\">\n        <XCUIElementTypeButton type=\"XCUIElementTypeButton\" value=\"1\" name=\"今天\" label=\"今天\" enabled=\"true\" visible=\"true\" x=\"2\" y=\"730\" width=\"90\" height=\"48\"/>\n        <XCUIElementTypeButton type=\"XCUIElementTypeButton\" name=\"健康数据\" label=\"健康数据\" enabled=\"true\" visible=\"true\" x=\"96\" y=\"730\" width=\"90\" height=\"48\"/>\n        <XCUIElementTypeButton type=\"XCUIElementTypeButton\" name=\"数据来源\" label=\"数据来源\" enabled=\"true\" visible=\"true\" x=\"190\" y=\"730\" width=\"89\" height=\"48\"/>\n        <XCUIElementTypeButton type=\"XCUIElementTypeButton\" name=\"医疗急救卡\" label=\"医疗急救卡\" enabled=\"true\" visible=\"true\" x=\"283\" y=\"730\" width=\"90\" height=\"48\"/>\n      </XCUIElementTypeTabBar>\n    </XCUIElementTypeOther>\n  </XCUIElementTypeWindow>\n  <XCUIElementTypeWindow type=\"XCUIElementTypeWindow\" enabled=\"true\" visible=\"false\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n    <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"false\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n      <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"false\" x=\"0\" y=\"812\" width=\"375\" height=\"233\"/>\n    </XCUIElementTypeOther>\n  </XCUIElementTypeWindow>\n  <XCUIElementTypeWindow type=\"XCUIElementTypeWindow\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"812\">\n    <XCUIElementTypeStatusBar type=\"XCUIElementTypeStatusBar\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"44\">\n      <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"44\">\n        <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"0\" y=\"0\" width=\"375\" height=\"44\">\n          <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"14\" y=\"0\" width=\"172\" height=\"29\"/>\n          <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"true\" x=\"190\" y=\"0\" width=\"171\" height=\"29\"/>\n          <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"false\" x=\"14\" y=\"10\" width=\"68\" height=\"19\">\n            <XCUIElementTypeStaticText type=\"XCUIElementTypeStaticText\" value=\"10:56\" name=\"10:56\" label=\"10:56\" enabled=\"true\" visible=\"false\" x=\"27\" y=\"14\" width=\"42\" height=\"18\"/>\n          </XCUIElementTypeOther>\n          <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"false\" x=\"21\" y=\"8\" width=\"56\" height=\"22\"/>\n          <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"false\" x=\"293\" y=\"10\" width=\"68\" height=\"19\">\n            <XCUIElementTypeOther type=\"XCUIElementTypeOther\" name=\"信号强度：4（共 4 格）\" label=\"信号强度：4（共 4 格）\" enabled=\"true\" visible=\"false\" x=\"293\" y=\"17\" width=\"18\" height=\"12\"/>\n            <XCUIElementTypeStaticText type=\"XCUIElementTypeStaticText\" value=\"4G\" name=\"4G\" label=\"4G\" enabled=\"true\" visible=\"false\" x=\"315\" y=\"16\" width=\"17\" height=\"15\"/>\n            <XCUIElementTypeOther type=\"XCUIElementTypeOther\" value=\"正在充电\" name=\"电池电量：68%\" label=\"电池电量：68%\" enabled=\"true\" visible=\"false\" x=\"336\" y=\"17\" width=\"25\" height=\"12\"/>\n          </XCUIElementTypeOther>\n          <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"false\" x=\"293\" y=\"10\" width=\"68\" height=\"19\"/>\n        </XCUIElementTypeOther>\n        <XCUIElementTypeOther type=\"XCUIElementTypeOther\" enabled=\"true\" visible=\"false\" x=\"8\" y=\"24\" width=\"82\" height=\"14\"/>\n      </XCUIElementTypeOther>\n    </XCUIElementTypeStatusBar>\n  </XCUIElementTypeWindow>\n</XCUIElementTypeApplication>\n"

    import json
    import requests
    jsonStr = requests.get("http://10.254.51.239:8100/source?format=json")
    jsonStr = jsonStr.text

    jsonObj = json.loads(jsonStr)

    dump = ios_dump_xml(XML, (375, 812))

    dumpJson = ios_dump_json(jsonObj["value"], (375, 812))
    pprint(dumpJson)



# coding=utf-8

import time
import unittest

from poco.drivers.cocosjs import CocosJsPoco


class SimpleTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.poco = CocosJsPoco("ws://localhost:15003")

    @classmethod
    def tearDownClass(cls):
        time.sleep(1)

    def test_dump(self):
        print(self.poco.agent.hierarchy.dump())



# -*- coding: utf-8 -*-
# @Author: gzliuxin
# @Email:  gzliuxin@corp.netease.com
# @Date:   2017-07-11 14:34:46

from hrpc.exceptions import TransportDisconnected

from poco.sdk.interfaces.hierarchy import HierarchyInterface
from poco.utils.hrpc.utils import transform_node_has_been_removed_exception
from poco.utils.retry import retries_when


class RemotePocoHierarchy(HierarchyInterface):
    def __init__(self, dumper, selector, attributor):
        super(RemotePocoHierarchy, self).__init__()
        self.dumper = dumper
        self.selector = selector
        self.attributor = attributor

    # node/hierarchy interface
    @retries_when(TransportDisconnected, delay=3.0)
    @transform_node_has_been_removed_exception
    def getAttr(self, nodes, name):
        return self.attributor.getAttr(nodes, name)

    @retries_when(TransportDisconnected, delay=3.0)
    @transform_node_has_been_removed_exception
    def setAttr(self, nodes, name, value):
        return self.attributor.setAttr(nodes, name, value)

    @retries_when(TransportDisconnected, delay=3.0)
    def select(self, query, multiple=False):
        return self.selector.select(query, multiple)

    @retries_when(TransportDisconnected, delay=3.0)
    def dump(self):
        return self.dumper.dumpHierarchy()



from __future__ import absolute_import

import copy
import json
import logging
import time
from uuid import uuid4

from flask import Blueprint, request, Response

from ..exceptions import JSONRPCInvalidRequestException
from ..jsonrpc import JSONRPCRequest
from ..manager import JSONRPCResponseManager
from ..utils import DatetimeDecimalEncoder
from ..dispatcher import Dispatcher


logger = logging.getLogger(__name__)


class JSONRPCAPI(object):
    def __init__(self, dispatcher=None, check_content_type=True):
        """

        :param dispatcher: methods dispatcher
        :param check_content_type: if True - content-type must be
            "application/json"
        :return:

        """
        self.dispatcher = dispatcher if dispatcher is not None \
            else Dispatcher()
        self.check_content_type = check_content_type

    def as_blueprint(self, name=None):
        blueprint = Blueprint(name if name else str(uuid4()), __name__)
        blueprint.add_url_rule(
            '/', view_func=self.jsonrpc, methods=['POST'])
        blueprint.add_url_rule(
            '/map', view_func=self.jsonrpc_map, methods=['GET'])
        return blueprint

    def as_view(self):
        return self.jsonrpc

    def jsonrpc(self):
        request_str = self._get_request_str()
        try:
            jsonrpc_request = JSONRPCRequest.from_json(request_str)
        except (TypeError, ValueError, JSONRPCInvalidRequestException):
            response = JSONRPCResponseManager.handle(
                request_str, self.dispatcher)
        else:
            jsonrpc_request.params = jsonrpc_request.params or {}
            jsonrpc_request_params = copy.copy(jsonrpc_request.params)
            t1 = time.time()
            response = JSONRPCResponseManager.handle_request(
                jsonrpc_request, self.dispatcher)
            t2 = time.time()
            logger.info('{0}({1}) {2:.2f} sec'.format(
                jsonrpc_request.method, jsonrpc_request_params, t2 - t1))

        if response:
            response.serialize = self._serialize
            response = response.json

        return Response(response, content_type="application/json")

    def jsonrpc_map(self):
        """ Map of json-rpc available calls.

        :return str:

        """
        result = "<h1>JSON-RPC map</h1><pre>{0}</pre>".format("\n\n".join([
            "{0}: {1}".format(fname, f.__doc__)
            for fname, f in self.dispatcher.items()
        ]))
        return Response(result)

    def _get_request_str(self):
        if self.check_content_type or request.data:
            return request.data
        return list(request.form.keys())[0]

    @staticmethod
    def _serialize(s):
        return json.dumps(s, cls=DatetimeDecimalEncoder)


api = JSONRPCAPI()



# coding=utf-8

import fnmatch
import atomac
import AppKit
from atomac import _a11y
import Quartz

pressID = [None, Quartz.kCGEventLeftMouseDown,
           Quartz.kCGEventRightMouseDown, Quartz.kCGEventOtherMouseDown]
releaseID = [None, Quartz.kCGEventLeftMouseUp,
             Quartz.kCGEventRightMouseUp, Quartz.kCGEventOtherMouseUp]


class OSXFunc(object):

    @classmethod
    def getRunningApps(cls):
        ws = AppKit.NSWorkspace.sharedWorkspace()
        apps = ws.runningApplications()
        return apps

    @classmethod
    def getAppRefByPid(cls, pid):
        return _a11y.getAppRefByPid(atomac.AXClasses.NativeUIElement, pid)

    @classmethod
    def getAppRefByBundleId(cls, bundleId):
        """
        Get the top level element for the application with the specified
        bundle ID, such as com.vmware.fusion.
        """
        ra = AppKit.NSRunningApplication
        # return value (apps) is always an array. if there is a match it will
        # have an item, otherwise it won't.
        apps = ra.runningApplicationsWithBundleIdentifier_(bundleId)
        if len(apps) == 0:
            raise ValueError(('Specified bundle ID not found in '
                            'running apps: %s' % bundleId))
        pid = apps[0].processIdentifier()
        return cls.getAppRefByPid(pid)   

    @classmethod
    def getAppRefByLocalizedName(cls, name):
        apps = cls.getRunningApps()
        for app in apps:
            if fnmatch.fnmatch(app.localizedName(), name):
                pid = app.processIdentifier()
                return cls.getAppRefByPid(pid)
        raise ValueError('Specified application not found in running apps.')

    @staticmethod
    def press(x, y, button=1):
        event = Quartz.CGEventCreateMouseEvent(None, pressID[button], (x, y), button - 1)
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, event)

    @staticmethod
    def release(x, y, button=1):
        event = Quartz.CGEventCreateMouseEvent(None, releaseID[button], (x, y), button - 1)
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, event)

    @staticmethod
    def click(x, y, button=1):
        theEvent = Quartz.CGEventCreateMouseEvent(None, pressID[button], (x, y), button - 1)
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, theEvent)  
        Quartz.CGEventSetType(theEvent, Quartz.kCGEventLeftMouseUp)  
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, theEvent)  

    @staticmethod
    def rclick(x, y, button=2):
        theEvent = Quartz.CGEventCreateMouseEvent(None, pressID[button], (x, y), button - 1)
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, theEvent)  
        Quartz.CGEventSetType(theEvent, releaseID[button])  
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, theEvent)  

    @staticmethod
    def doubleclick(x, y, button=1):
        theEvent = Quartz.CGEventCreateMouseEvent(None, pressID[button], (x, y), button - 1)
        Quartz.CGEventSetIntegerValueField(theEvent, Quartz.kCGMouseEventClickState, 2)  
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, theEvent)  
        Quartz.CGEventSetType(theEvent, Quartz.kCGEventLeftMouseUp)  
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, theEvent)  
        Quartz.CGEventSetType(theEvent, Quartz.kCGEventLeftMouseDown) 
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, theEvent)
        Quartz.CGEventSetType(theEvent, Quartz.kCGEventLeftMouseUp)
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, theEvent)

    @staticmethod
    def move(x, y):
        move = Quartz.CGEventCreateMouseEvent(None, Quartz.kCGEventMouseMoved, (x, y), 0)
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, move)

    @staticmethod
    def drag(x, y):
        drag = Quartz.CGEventCreateMouseEvent(None, Quartz.kCGEventLeftMouseDragged, (x, y), 0)
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, drag)

    @staticmethod
    def scroll(vertical=None, horizontal=None, depth=None):
        # Local submethod for generating Mac scroll events in one axis at a time
        def scroll_event(y_move=0, x_move=0, z_move=0, n=1):
            for _ in range(abs(n)):
                scrollWheelEvent = Quartz.CGEventCreateScrollWheelEvent(
                    None,  # No source
                    Quartz.kCGScrollEventUnitLine,  # Unit of measurement is lines
                    3,  # Number of wheels(dimensions)
                    y_move,
                    x_move,
                    z_move)
                Quartz.CGEventPost(Quartz.kCGHIDEventTap, scrollWheelEvent)

        # Execute vertical then horizontal then depth scrolling events
        if vertical is not None:
            vertical = int(vertical)
            if vertical == 0:   # Do nothing with 0 distance
                pass
            elif vertical > 0:  # Scroll up if positive
                scroll_event(y_move=1, n=vertical)
            else:  # Scroll down if negative
                scroll_event(y_move=-1, n=abs(vertical))
        if horizontal is not None:
            horizontal = int(horizontal)
            if horizontal == 0:  # Do nothing with 0 distance
                pass
            elif horizontal > 0:  # Scroll right if positive
                scroll_event(x_move=1, n=horizontal)
            else:  # Scroll left if negative
                scroll_event(x_move=-1, n=abs(horizontal))
        if depth is not None:
            depth = int(depth)
            if depth == 0:  # Do nothing with 0 distance
                pass
            elif vertical > 0:  # Scroll "out" if positive
                scroll_event(z_move=1, n=depth)
            else:  # Scroll "in" if negative
                scroll_event(z_move=-1, n=abs(depth))



# _*_ coding:UTF-8 _*_
import socket


DEFAULT_TIMEOUT = 5
DEFAULT_SIZE = 4096


class Client(object):
    """safe and exact recv & send"""
    def __init__(self, address, on_connect=None, on_close=None):
        """address is (host, port) tuple"""
        self.address = address
        self.on_connect = on_connect
        self.on_close = on_close
        self.sock = None
        self.buf = b""

    def connect(self):
        # create a new socket every time
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.settimeout(DEFAULT_TIMEOUT)
        self.sock.connect(self.address)
        self._handle_connect()

    def send(self, msg):
        totalsent = 0
        while totalsent < len(msg):
            sent = self.sock.send(msg[totalsent:])
            if sent == 0:
                self._handle_close()
                raise socket.error("socket connection broken")
            totalsent += sent

    def recv(self, size=DEFAULT_SIZE):
        trunk = self.sock.recv(size)
        if trunk == b"":
            self._handle_close()
            raise socket.error("socket connection broken")
        return trunk

    def recv_all(self, size):
        while len(self.buf) < size:
            trunk = self.recv(min(size-len(self.buf), DEFAULT_SIZE))
            self.buf += trunk
        ret, self.buf = self.buf[:size], self.buf[size:]
        return ret

    def settimeout(self, timeout):
        self.sock.settimeout(timeout)

    def recv_nonblocking(self, size):
        self.sock.settimeout(0)
        try:
            ret = self.recv(size)
        except socket.error as e:
            # 10035 no data when nonblocking
            if e.args[0] == 10035:  # errno.EWOULDBLOCK: errno is not always right
                ret = None
            # 10053 connection abort by client
            # 10054 connection reset by peer
            elif e.args[0] in [10053, 10054]:  # errno.ECONNABORTED:
                raise
            else:
                raise
        return ret

    def close(self):
        self.sock.shutdown(socket.SHUT_RDWR)
        self.sock.close()
        self._handle_close()

    def _handle_connect(self):
        if callable(self.on_connect):
            self.on_connect()

    def _handle_close(self):
        if callable(self.on_close):
            self.on_close()



# coding=utf-8

from typing import Iterable, Iterator, Text, Tuple, Union, List, Dict, Any, NoReturn

from poco.pocofw import Poco
from poco.gesture import PendingGestureAction
from poco.sdk.AbstractNode import AbstractNode

class UIObjectProxy(object):

    def __init__(self, poco, name=None, **attrs) -> UIObjectProxy:
        self.query = ...                    # type: (str, Tuple)
        self.poco = ...                     # type: Poco
        self._query_multiple = ...          # type: bool
        self._evaluated = ...               # type: bool
        self._nodes = ...                   # type: Union[type(None), List[AbstractNode]]
        self._nodes_proxy_is_list = ...     # type: bool
        self._sorted_children = ...         # type: Union[type(None), (UIObjectProxy, (float, float))]
        self._focus = ...                   # type: Union[type(None), (float, float)]

    def child(self, name: Text=None, **attrs) -> UIObjectProxy:
        ...

    def children(self) -> UIObjectProxy:
        ...

    def offspring(self, name: Text=None, **attrs) -> UIObjectProxy:
        ...

    def sibling(self, name: Text=None, **attrs) -> UIObjectProxy:
        ...

    def parent(self) -> UIObjectProxy:
        ...

    def __getitem__(self, item: int) -> UIObjectProxy:
        ...

    def __iter__(self) -> Iterator[UIObjectProxy]:
        ...

    def click(self, focus: (float, float)=None, sleep_interval: float=None) -> bool:
        ...

    def long_click(self, duration: float=2.0) -> bool:
        ...

    def swipe(self, direction: (float, float), focus: (float, float)=None, duration=0.5) -> bool:
        ...

    def drag_to(self, target: UIObjectProxy, duration: float=2.0) -> bool:
        ...

    def scroll(self, direction: Text='vertical', percent: float=0.6, duration: float=2.0) -> bool:
        ...

    def pinch(self, direction: Text='in', percent: float=0.6, duration: float=2.0, dead_zone: float=0.1) -> bool:
        ...

    def start_gesture(self) -> PendingGestureAction:
        ...

    def focus(self, f: (float, float)) -> UIObjectProxy:
        ...

    def get_position(self, focus: (float, float)=None) -> (float, float):
        ...

    def get_text(self) -> Text:
        ...

    def set_text(self, text: Text) -> bool:
        ...

    def get_name(self) -> Text: ...
    def get_size(self) -> (float, float): ...
    def get_bounds(self) -> (float, float, float, float): ...

    @property
    def nodes(self) -> List[AbstractNode]: ...

    def wait(self, timeout: float=3.0) -> UIObjectProxy:
        ...

    def wait_for_appearance(self, timeout: float=120) -> NoReturn: ...
    def wait_for_disappearance(self, timeout: float=120) -> NoReturn: ...

    def attr(self, name: Text) -> Any:
        ...

    def setattr(self, name: Text, val: Any) -> bool:
        ...

    def exists(self) -> bool:
        ...

    def invalidate(self) -> NoReturn:
        ...

    def _do_query(self, multiple: bool=True, refresh: bool=False) -> Union[type(None), List[AbstractNode]]:
        ...

    def _direction_vector_of(self, dir_: (float, float)) -> (float, float):
        ...



# coding=utf-8

from typing import Iterable, Iterator, Text, Tuple, Union, List, Dict, Any, NoReturn

from poco.pocofw import Poco
from poco.sdk.interfaces.hierarchy import HierarchyInterface
from poco.sdk.interfaces.input import InputInterface
from poco.sdk.interfaces.screen import ScreenInterface
from poco.sdk.interfaces.command import CommandInterface


class PocoAgent(object):
    def __init__(self,
                 hierarchy: HierarchyInterface,
                 input: InputInterface,
                 screen: ScreenInterface,
                 command: CommandInterface=None):
        self.hierarchy = ...    # type: HierarchyInterface
        self.input = ...        # type: InputInterface
        self.screen = ...       # type: ScreenInterface
        self.command = ...      # type: CommandInterface
        self._driver = ...      # type: Poco

    def get_sdk_version(self) -> Text:
        ...

    @property
    def rpc(self) -> Any:
        ...

    def on_bind_driver(self, driver: Poco) -> NoReturn:
        ...

    @property
    def driver(self) -> Poco:
        ...



# coding=utf-8

import time
import threading

from poco.sdk.std.rpc.controller import StdRpcEndpointController
from poco.sdk.std.rpc.reactor import StdRpcReactor
from poco.utils.net.transport.tcp import TcpSocket
from poco.utils.net.stdbroker import StdBroker


def Dump(arg):
    return 'this is Dump ' + arg


if __name__ == '__main__':
    # broker = StdBroker('tcp://*:15003', 'tcp://*:15004')

    reactor = StdRpcReactor()
    reactor.register('Dump', Dump)

    responser = TcpSocket()
    responser.connect(('localhost', 15003))

    rpc_responser = StdRpcEndpointController(responser, reactor)
    t = threading.Thread(target=rpc_responser.serve_forever)
    t.daemon = True
    t.start()

    requester = TcpSocket()
    requester.connect(('localhost', 15004))
    rpc_requester = StdRpcEndpointController(requester, StdRpcReactor())
    t = threading.Thread(target=rpc_requester.serve_forever)
    t.daemon = True
    t.start()

    requester = TcpSocket()
    requester.connect(('localhost', 15004))
    rpc_requester2 = StdRpcEndpointController(requester, StdRpcReactor())
    t = threading.Thread(target=rpc_requester2.serve_forever)
    t.daemon = True
    t.start()


    print(1111, rpc_requester.call('Dump', '111111'))
    print(2222, rpc_requester2.call('Dump', '222222'))

    while True:
        time.sleep(1)



# coding=utf-8

from poco.sdk.interfaces.input import InputInterface
from poco.utils.simplerpc.utils import sync_wrapper


class StdInput(InputInterface):
    def __init__(self, client):
        super(StdInput, self).__init__()
        self.client = client
        # 根据需要修改构造函数的签名
        # 并修改对应的调用处

    @sync_wrapper
    def click(self, x, y):
        return self.client.call("Click", x, y)

    @sync_wrapper
    def swipe(self, x1, y1, x2, y2, duration):
        return self.client.call("Swipe", x1, y1, x2, y2, duration)

    @sync_wrapper
    def longClick(self, x, y, duration):
        return self.client.call("LongClick", x, y, duration)

    @sync_wrapper
    def keyevent(self, keycode):
        return self.client.call("KeyEvent", keycode)

    @sync_wrapper
    def scroll(self, direction='vertical', percent=1, duration=2.0):
        return self.client.call("Scroll", direction, percent, duration)

    @sync_wrapper
    def rclick(self, x, y):
        return self.client.call("RClick", x, y)

    @sync_wrapper
    def double_click(self, x, y):
        return self.client.call("DoubleClick", x, y)



# coding=utf-8
__author__ = 'lxn3032'


import os
import requests
import time
import warnings
import threading
import atexit

from airtest.core.android.ime import YosemiteIme
from airtest.core.error import AdbShellError, AirtestError

from hrpc.client import RpcClient
from hrpc.transport.http import HttpTransport
from poco.pocofw import Poco
from poco.agent import PocoAgent
from poco.sdk.Attributor import Attributor
from poco.sdk.interfaces.screen import ScreenInterface
from poco.utils.hrpc.hierarchy import RemotePocoHierarchy
from poco.utils.airtest.input import AirtestInput
from poco.utils import six
from poco.utils.device import default_device
from poco.drivers.android.utils.installation import install, uninstall

__all__ = ['AndroidUiautomationPoco', 'AndroidUiautomationHelper']
this_dir = os.path.dirname(os.path.realpath(__file__))
PocoServicePackage = 'com.netease.open.pocoservice'
PocoServicePackageTest = 'com.netease.open.pocoservice.test'
UiAutomatorPackage = 'com.github.uiautomator'


class AndroidRpcClient(RpcClient):
    def __init__(self, endpoint):
        self.endpoint = endpoint
        super(AndroidRpcClient, self).__init__(HttpTransport)

    def initialize_transport(self):
        return HttpTransport(self.endpoint, self)


# deprecated
class AttributorWrapper(Attributor):
    """
    部分手机上仍不支持Accessibility.ACTION_SET_TEXT，使用YosemiteIme还是兼容性最好的方案
    这个class会hook住set_text，然后改用ime的text方法
    """

    def __init__(self, remote, ime):
        self.remote = remote
        self.ime = ime

    def getAttr(self, node, attrName):
        return self.remote.getAttr(node, attrName)

    def setAttr(self, node, attrName, attrVal):
        if attrName == 'text' and attrVal != '':
            # 先清除了再设置，虽然这样不如直接用ime的方法好，但是也能凑合用着
            current_val = self.remote.getAttr(node, 'text')
            if current_val:
                self.remote.setAttr(node, 'text', '')
            self.ime.text(attrVal)
        else:
            self.remote.setAttr(node, attrName, attrVal)


class ScreenWrapper(ScreenInterface):
    def __init__(self, screen):
        super(ScreenWrapper, self).__init__()
        self.screen = screen

    def getScreen(self, width):
        # Android上PocoService的实现为仅返回b64编码的图像，格式固定位jpg
        b64img = self.screen.getScreen(width)
        return b64img, 'jpg'

    def getPortSize(self):
        return self.screen.getPortSize()


class AndroidPocoAgent(PocoAgent):
    def __init__(self, endpoint, ime, use_airtest_input=False):
        self.client = AndroidRpcClient(endpoint)
        remote_poco = self.client.remote('poco-uiautomation-framework')
        dumper = remote_poco.dumper
        selector = remote_poco.selector
        attributor = remote_poco.attributor
        hierarchy = RemotePocoHierarchy(dumper, selector, attributor)

        if use_airtest_input:
            inputer = AirtestInput()
        else:
            inputer = remote_poco.inputer
        super(AndroidPocoAgent, self).__init__(hierarchy, inputer, ScreenWrapper(remote_poco.screen), None)


class KeepRunningInstrumentationThread(threading.Thread):
    """Keep pocoservice running"""

    def __init__(self, poco, port_to_ping):
        super(KeepRunningInstrumentationThread, self).__init__()
        self._stop_event = threading.Event()
        self.poco = poco
        self.port_to_ping = port_to_ping
        self.daemon = True

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()

    def run(self):
        while not self.stopped():
            if not self.stopped():
                self.poco._start_instrument(self.port_to_ping)  # 尝试重启
                time.sleep(1)


class AndroidUiautomationPoco(Poco):
    """
    Poco Android implementation for testing **Android native apps**.

    Args:
        device (:py:obj:`Device`): :py:obj:`airtest.core.device.Device` instance provided by ``airtest``. leave the 
         parameter default and the default device will be chosen. more details refer to ``airtest doc``
        using_proxy (:py:obj:`bool`): whether use adb forward to connect the Android device or not
        force_restart (:py:obj:`bool`): whether always restart the poco-service-demo running on Android device or not
        options: see :py:class:`poco.pocofw.Poco`

    Examples:
        The simplest way to initialize AndroidUiautomationPoco instance and no matter your device network status::

            from poco.drivers.android.uiautomation import AndroidUiautomationPoco

            poco = AndroidUiautomationPoco()
            poco('android:id/title').click()
            ...

    """

    def __init__(self, device=None, using_proxy=True, force_restart=False, use_airtest_input=False, **options):
        # 加这个参数为了不在最新的pocounit方案中每步都截图
        self.screenshot_each_action = True
        if options.get('screenshot_each_action') is False:
            self.screenshot_each_action = False

        self.device = device or default_device()

        self.adb_client = self.device.adb
        if using_proxy:
            self.device_ip = self.adb_client.host or "127.0.0.1"
        else:
            self.device_ip = self.device.get_ip_address()

        # save current top activity (@nullable)
        try:
            current_top_activity_package = self.device.get_top_activity_name()
        except AirtestError as e:
            # 在一些极端情况下，可能获取不到top activity的信息
            print(e)
            current_top_activity_package = None
        if current_top_activity_package is not None:
            current_top_activity_package = current_top_activity_package.split('/')[0]

        # install ime
        self.ime = YosemiteIme(self.adb_client)

        # install
        self._instrument_proc = None
        self._install_service()

        # forward
        self.forward_list = []
        if using_proxy:
            p0, _ = self.adb_client.setup_forward("tcp:10080")
            p1, _ = self.adb_client.setup_forward("tcp:10081")
            self.forward_list.extend(["tcp:%s" % p0, "tcp:%s" % p1])
        else:
            p0 = 10080
            p1 = 10081

        # start
        ready = self._start_instrument(p0, force_restart=force_restart)
        if not ready:
            # 之前启动失败就卸载重装，现在改为尝试kill进程或卸载uiautomator
            self._kill_uiautomator()
            ready = self._start_instrument(p0)

            if current_top_activity_package is not None:
                current_top_activity2 = self.device.get_top_activity_name()
                if current_top_activity2 is None or current_top_activity_package not in current_top_activity2:
                    self.device.start_app(current_top_activity_package, activity=True)

            if not ready:
                raise RuntimeError("unable to launch AndroidUiautomationPoco")
        if ready:
            # 首次启动成功后，在后台线程里监控这个进程的状态，保持让它不退出
            self._keep_running_thread = KeepRunningInstrumentationThread(self, p0)
            self._keep_running_thread.start()

        endpoint = "http://{}:{}".format(self.device_ip, p1)
        agent = AndroidPocoAgent(endpoint, self.ime, use_airtest_input)
        super(AndroidUiautomationPoco, self).__init__(agent, **options)

    def _install_service(self):
        updated = install(self.adb_client, os.path.join(this_dir, 'lib', 'pocoservice-debug.apk'))
        return updated

    def _is_running(self, package_name):
        """
        use ps |grep to check whether the process exists

        :param package_name: package name(e.g., com.github.uiautomator)
                            or regular expression(e.g., poco\|airtest\|uiautomator\|airbase)
        :return: pid or None
        """
        cmd = r' |echo $(grep -E {package_name})'.format(package_name=package_name)
        if self.device.sdk_version > 25:
            cmd = r'ps -A' + cmd
        else:
            cmd = r'ps' + cmd
        processes = self.adb_client.shell(cmd).splitlines()
        for ps in processes:
            if ps:
                ps = ps.split()
                return ps[1]
        return None

    def _start_instrument(self, port_to_ping, force_restart=False):
        if not force_restart:
            try:
                state = requests.get('http://{}:{}/uiautomation/connectionState'.format(self.device_ip, port_to_ping),
                                     timeout=10)
                state = state.json()
                if state.get('connected'):
                    # skip starting instrumentation if UiAutomation Service already connected.
                    return True
            except:
                pass

        if self._instrument_proc is not None:
            if self._instrument_proc.poll() is None:
                self._instrument_proc.kill()
            self._instrument_proc = None

        ready = False
        # self.adb_client.shell(['am', 'force-stop', PocoServicePackage])

        # 启动instrument之前，先把主类activity启动起来，不然instrumentation可能失败
        self.adb_client.shell('am start -n {}/.TestActivity'.format(PocoServicePackage))

        instrumentation_cmd = [
            'am', 'instrument', '-w', '-e', 'debug', 'false', '-e', 'class',
            '{}.InstrumentedTestAsLauncher'.format(PocoServicePackage),
            '{}/androidx.test.runner.AndroidJUnitRunner'.format(PocoServicePackage)]
        self._instrument_proc = self.adb_client.start_shell(instrumentation_cmd)

        def cleanup_proc(proc):
            def wrapped():
                try:
                    proc.kill()
                except:
                    pass
            return wrapped
        atexit.register(cleanup_proc(self._instrument_proc))

        time.sleep(2)
        for i in range(10):
            try:
                requests.get('http://{}:{}'.format(self.device_ip, port_to_ping), timeout=10)
                ready = True
                break
            except requests.exceptions.Timeout:
                break
            except requests.exceptions.ConnectionError:
                if self._instrument_proc.poll() is not None:
                    warnings.warn("[pocoservice.apk] instrumentation test server process is no longer alive")
                    stdout = self._instrument_proc.stdout.read()
                    stderr = self._instrument_proc.stderr.read()
                    print('[pocoservice.apk] stdout: {}'.format(stdout))
                    print('[pocoservice.apk] stderr: {}'.format(stderr))
                time.sleep(1)
                print("still waiting for uiautomation ready.")
                try:
                    self.adb_client.shell(
                        ['monkey', '-p', {PocoServicePackage}, '-c', 'android.intent.category.LAUNCHER', '1'])
                except Exception as e:
                    pass
                self.adb_client.shell('am start -n {}/.TestActivity'.format(PocoServicePackage))
                instrumentation_cmd = [
                    'am', 'instrument', '-w', '-e', 'debug', 'false', '-e', 'class',
                    '{}.InstrumentedTestAsLauncher'.format(PocoServicePackage),
                    '{}/androidx.test.runner.AndroidJUnitRunner'.format(PocoServicePackage)]
                self._instrument_proc = self.adb_client.start_shell(instrumentation_cmd)
                continue
        return ready

    def _kill_uiautomator(self):
        """
        poco-service无法与其他instrument启动的apk同时存在，因此在启动前，需要杀掉一些可能的进程：
        比如 io.appium.uiautomator2.server, com.github.uiautomator, com.netease.open.pocoservice等

        :return:
        """
        pid = self._is_running("uiautomator")
        if pid:
            warnings.warn('{} should not run together with "uiautomator". "uiautomator" will be killed.'
                          .format(self.__class__.__name__))
            self.adb_client.shell(['am', 'force-stop', PocoServicePackage])

            try:
                self.adb_client.shell(['kill', pid])
            except AdbShellError:
                # 没有root权限
                uninstall(self.adb_client, UiAutomatorPackage)

    def on_pre_action(self, action, ui, args):
        if self.screenshot_each_action:
            # airteset log用
            from airtest.core.api import snapshot
            msg = repr(ui)
            if not isinstance(msg, six.text_type):
                msg = msg.decode('utf-8')
            snapshot(msg=msg)

    def stop_running(self):
        print('[pocoservice.apk] stopping PocoService')
        self._keep_running_thread.stop()
        self._keep_running_thread.join(3)
        self.remove_forwards()
        self.adb_client.shell(['am', 'force-stop', PocoServicePackage])

    def remove_forwards(self):
        for p in self.forward_list:
            self.adb_client.remove_forward(p)
        self.forward_list = []


class AndroidUiautomationHelper(object):
    _nuis = {}

    @classmethod
    def get_instance(cls, device):
        """
        This is only a slot to store and get already initialized poco instance rather than initializing again. You can
        simply pass the ``current device instance`` provided by ``airtest`` to get the AndroidUiautomationPoco instance.
        If no such AndroidUiautomationPoco instance, a new instance will be created and stored. 

        Args:
            device (:py:obj:`airtest.core.device.Device`): more details refer to ``airtest doc``

        Returns:
            poco instance
        """

        if cls._nuis.get(device) is None:
            cls._nuis[device] = AndroidUiautomationPoco(device)
        return cls._nuis[device]



# coding=utf-8

from poco.utils.track import MotionTrack

__all__ = ['PendingGestureAction']


class PendingGestureAction(object):
    def __init__(self, pocoobj, uiproxy_or_pos):
        super(PendingGestureAction, self).__init__()
        self.pocoobj = pocoobj
        self.track = MotionTrack()
        if isinstance(uiproxy_or_pos, (list, tuple)):
            self.track.start(uiproxy_or_pos)
        else:
            self.track.start(uiproxy_or_pos.get_position())

    def hold(self, t):
        self.track.hold(t)
        return self

    def to(self, pos):
        if isinstance(pos, (list, tuple)):
            self.track.move(pos)
        else:
            uiobj = pos
            self.track.move(uiobj.get_position())
        return self

    def up(self):
        self.pocoobj.apply_motion_tracks([self.track])



# coding=utf-8

import base64
import json
import traceback
import time
import unittest

from poco.drivers.std import StdPoco
from poco.utils.simplerpc.utils import sync_wrapper, RemoteError
from airtest.core.api import connect_device


class TestStandardFunction(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # u3d game 默认5001
        # cocos2dx-lua 默认15004
        connect_device('Android:///')
        # connect_device('Windows:///?class_name=UnityWndClass&title_re=Unity.*')
        cls.poco = StdPoco(15004)

    @classmethod
    def tearDownClass(cls):
        time.sleep(1)

    def test_dump(self):
        h = self.poco.agent.hierarchy.dump()
        s = json.dumps(h, indent=4)
        print(s)
        self.assertNotIn('"visible": false', s)

    def test_dump_include_invisible_node(self):
        h = self.poco.agent.hierarchy.dumper.dumpHierarchy(onlyVisibleNode=False)  # .hierarchy是FrozenUIHierarchy
        s = json.dumps(h, indent=4)
        print(s)
        self.assertIn('"visible": false', s)

    def test_getSdkVersion(self):
        print(self.poco.agent.get_sdk_version())

    def test_no_such_rpc_method(self):
        @sync_wrapper
        def wrapped(*args, **kwargs):
            return self.poco.agent.c.call('no_such_method')

        with self.assertRaises(RemoteError):
            wrapped()
        try:
            wrapped()
        except:
            traceback.print_exc()

    def test_get_screen(self):
        b64img, fmt = self.poco.snapshot()
        with open('screen.{}'.format(fmt), 'wb') as f:
            f.write(base64.b64decode(b64img))

    def test_get_screen_size(self):
        print(self.poco.get_screen_size())

    def test_motion_events(self):
        ui = self.poco()[0]
        ui.click()
        time.sleep(0.5)
        ui.long_click()
        time.sleep(0.5)
        ui.swipe([0.1, 0.1])

    def test_set_text(self):
        textval = 'hello 中国!'
        node = self.poco(typeMatches='TextField|InputField|EditBox')
        node.set_text(textval)
        node.invalidate()
        actualVal = node.get_text() or node.offspring(text=textval).get_text()
        print(repr(actualVal))
        self.assertEqual(actualVal, textval)

    def test_clear_text(self):
        node = self.poco(typeMatches='TextField|InputField|EditBox')
        node.set_text('val2333')
        node.set_text('')
        node.invalidate()
        actual_val = node.get_text()
        if actual_val is None:
            actual_val = node.offspring(text='').get_text()
        self.assertEqual(actual_val, '')

    def test_instanceId(self):
        for n in self.poco():
            instance_id = n.attr('_instanceId')
            if instance_id:
                print(instance_id)



import logging


def init_logging():
    logger = logging.getLogger("poco")
    logger.setLevel(logging.DEBUG)
    error_filter = logging.Filter()
    error_filter.filter = lambda record: record.levelno == logging.ERROR

    if not logging.getLogger().hasHandlers():
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            fmt='[%(asctime)s][%(levelname)s]<%(name)s> %(message)s',
            datefmt='%H:%M:%S'
        )
        handler.setFormatter(formatter)
        handler.addFilter(error_filter)
        logger.addHandler(handler)
    else:
        for handler in logging.getLogger().handlers:
            handler.addFilter(error_filter)



init_logging()

def get_logger(name):
    logger = logging.getLogger(name)
    return logger



# coding=utf-8

import time
import base64
import zlib
import win32con
import win32gui
import re
import operator
import uiautomation as UIAuto
from poco.sdk.std.rpc.controller import StdRpcEndpointController
from poco.sdk.std.rpc.reactor import StdRpcReactor
from poco.utils.net.transport.tcp import TcpSocket
from poco.drivers.windows.sdk.WindowsUIDumper import WindowsUIDumper
from poco.sdk.exceptions import UnableToSetAttributeException, NonuniqueSurfaceException, InvalidSurfaceException
from poco.utils.six import string_types, PY2
from poco.utils.six.moves import reduce

DEFAULT_PORT = 15004
DEFAULT_ADDR = ('0.0.0.0', DEFAULT_PORT)


class PocoSDKWindows(object):

    def __init__(self, addr=DEFAULT_ADDR):
        self.reactor = None
        self.addr = addr
        self.running = False
        UIAuto.OPERATION_WAIT_TIME = 0.05  # make operation faster
        self.root = None

    def Dump(self, _):
        res = WindowsUIDumper(self.root).dumpHierarchy()
        return res

    def SetText(self, id, val2):
        control = UIAuto.ControlFromHandle(id)
        if not control or not isinstance(val2, string_types):
            raise UnableToSetAttributeException("text", control)
        else:
            control.SetValue(val2)

    def GetSDKVersion(self):
        return '0.0.1'

    def GetDebugProfilingData(self):
        return {}

    def GetScreenSize(self):
        Width = self.root.BoundingRectangle[2] - self.root.BoundingRectangle[0]
        Height = self.root.BoundingRectangle[3] - self.root.BoundingRectangle[1]
        return [Width, Height]

    def GetWindowRect(self):
        return self.root.BoundingRectangle

    def JudgeSize(self):
        self.SetForeground()  # 先打开窗口再获取大小，否则最小化的窗口获取到的大小会为0
        size = self.GetScreenSize()
        if size[0] == 0 or size[1] == 0:
            raise InvalidSurfaceException(self, "You may have minimized or closed your window or the window is too small!")

    def Screenshot(self, width):
        self.JudgeSize()
        self.root.ToBitmap().ToFile('Screenshot.bmp')
        f = open(r'Screenshot.bmp', 'rb')
        deflated = zlib.compress(f.read())  # 压缩图片
        ls_f = base64.b64encode(deflated)
        f.close()
        return [ls_f, "bmp.deflate"]

        # self.root.ToBitmap().ToFile('Screenshot.bmp')
        # f = open(r'Screenshot.bmp', 'rb')
        # ls_f = base64.b64encode(f.read())
        # f.close()
        # return [ls_f, "bmp"]

    def Click(self, x, y):
        self.JudgeSize()
        self.root.Click(x, y)
        return True
    
    def RClick(self, x, y):
        self.JudgeSize()
        self.root.RightClick(x, y)
        return True

    def DoubleClick(self, x, y):
        self.JudgeSize()
        self.root.DoubleClick(x, y)
        return True

    def Swipe(self, x1, y1, x2, y2, duration, **kwargs):
        self.JudgeSize()
        Left = self.root.BoundingRectangle[0]
        Top = self.root.BoundingRectangle[1]
        Width = self.root.BoundingRectangle[2] - self.root.BoundingRectangle[0]
        Height = self.root.BoundingRectangle[3] - self.root.BoundingRectangle[1]
        x1 = int(Left + Width * x1)  # 比例换算
        y1 = int(Top + Height * y1)
        x2 = int(Left + Width * x2)
        y2 = int(Top + Height * y2)

        UIAuto.MAX_MOVE_SECOND = duration * 10  # 同步到跟UIAutomation库的时间设定一样
        UIAuto.DragDrop(int(x1), int(y1), int(x2), int(y2))
        return True

    def LongClick(self, x, y, duration, **kwargs):
        self.JudgeSize()
        Left = self.root.BoundingRectangle[0]
        Top = self.root.BoundingRectangle[1]
        Width = self.root.BoundingRectangle[2] - self.root.BoundingRectangle[0]
        Height = self.root.BoundingRectangle[3] - self.root.BoundingRectangle[1]
        x = Left + Width * x
        y = Top + Height * y

        UIAuto.MAX_MOVE_SECOND = duration * 10
        UIAuto.DragDrop(int(x), int(y), int(x), int(y))
        return True

    def Scroll(self, direction, percent, duration):
        if direction not in ('vertical', 'horizontal'):
            return False

        if direction == 'horizontal':
            return False
        
        self.JudgeSize()
        x = 0.5  # 先把鼠标移到窗口中间，这样才能保证滚动的是这个窗口。
        y = 0.5
        steps = percent
        Left = self.root.BoundingRectangle[0]
        Top = self.root.BoundingRectangle[1]
        Width = self.root.BoundingRectangle[2] - self.root.BoundingRectangle[0]
        Height = self.root.BoundingRectangle[3] - self.root.BoundingRectangle[1]
        x = Left + Width * x
        y = Top + Height * y
        x = int(x)
        y = int(y)
        UIAuto.MoveTo(x, y)
        interval = float(duration - 0.3 * steps) / (abs(steps) + 1)  # 实现滚动时间
        if interval <= 0:
            interval = 0.1
        if steps < 0:
            for i in range(0, abs(steps)):
                time.sleep(interval)
                UIAuto.WheelUp(1)
        else:
            for i in range(0, abs(steps)):
                time.sleep(interval)
                UIAuto.WheelDown(1)
        return True

    def KeyEvent(self, keycode):
        self.JudgeSize()
        UIAuto.SendKeys(keycode)
        return True

    def SetForeground(self):
        win32gui.ShowWindow(self.root.Handle, win32con.SW_SHOWNORMAL)  # 先把窗口取消最小化
        UIAuto.Win32API.SetForegroundWindow(self.root.Handle)  # 再把窗口设为前台，方便点击和截图
        return True

    def EnumWindows(self):
        hWndList = []  # 枚举所有窗口，并把有效窗口handle保存在hwndlist里

        def foo(hwnd, mouse):
            if win32gui.IsWindow(hwnd):
                hWndList.append(hwnd)
        win32gui.EnumWindows(foo, 0)
        return hWndList

    def ConnectWindowsByTitle(self, title):
        hn = set()  # 匹配窗口的集合，把所有标题匹配上的窗口handle都保存在这个集合里
        hWndList = self.EnumWindows()
        for handle in hWndList:
            title_temp = win32gui.GetWindowText(handle)
            if PY2:
                title_temp = title_temp.decode("gbk")  # py2要解码成GBK，WindowsAPI中文返回的一般都是GBK
            if title == title_temp:
                hn.add(handle)
        if len(hn) == 0:
            return -1
        return hn

    def ConnectWindowsByTitleRe(self, title_re):
        hn = set()  # 匹配窗口的集合，把所有标题（正则表达式）匹配上的窗口handle都保存在这个集合里
        hWndList = self.EnumWindows()
        for handle in hWndList:
            title = win32gui.GetWindowText(handle)
            if PY2:
                title = title.decode("gbk")
            if re.match(title_re, title):
                hn.add(handle)
        if len(hn) == 0:
            return -1
        return hn

    def ConnectWindowsByHandle(self, handle):
        hn = set()  # 匹配窗口的集合，把所有handle匹配上的窗口handle都保存在这个集合里
        hWndList = self.EnumWindows()
        for handle_temp in hWndList:
            if int(handle_temp) == int(handle):
                hn.add(handle)
                break
        if len(hn) == 0:
            return -1
        return hn

    def ConnectWindow(self, selector):
        
        # 目前来说，如下处理，以后添加更多的参数后需修改代码逻辑
        argunums = 0
        if 'handle' in selector:
            argunums += 1
        if 'title' in selector:
            argunums += 1
        if 'title_re' in selector:
            argunums += 1
        
        if argunums == 0:
            raise ValueError("Expect handle or title, got none")
        elif argunums != 1:
            raise ValueError("Too many arguments, only need handle or title or title_re")

        handleSetList = []
        if 'title' in selector:
            handleSetList.append(self.ConnectWindowsByTitle(selector['title']))
        if 'handle' in selector:
            handleSetList.append(self.ConnectWindowsByHandle(selector['handle']))
        if "title_re" in selector:
            handleSetList.append(self.ConnectWindowsByTitleRe(selector['title_re']))

        while -1 in handleSetList:
            handleSetList.remove(-1)  # 有些参数没有提供会返回-1.把所有的-1去掉

        if len(handleSetList) is 0:
            raise InvalidSurfaceException(selector, "Can't find any windows by the given parameter")

        handleSet = reduce(operator.__and__, handleSetList)  # 提供了多个参数来确定唯一一个窗口，所以要做交集，取得唯一匹配的窗口

        if len(handleSet) == 0:
            raise InvalidSurfaceException(selector, "Can't find any windows by the given parameter")
        elif len(handleSet) != 1:
            raise NonuniqueSurfaceException(selector)
        else:
            hn = handleSet.pop()  # 取得那个唯一的窗口
            self.root = UIAuto.ControlFromHandle(hn)
            if self.root is None:
                raise InvalidSurfaceException(selector, "Can't find any windows by the given parameter")
            self.SetForeground()

    def run(self):
        self.reactor = StdRpcReactor()
        self.reactor.register('Dump', self.Dump)  # 注册各种函数
        self.reactor.register('SetText', self.SetText)
        self.reactor.register('GetSDKVersion', self.GetSDKVersion)
        self.reactor.register('GetDebugProfilingData', self.GetDebugProfilingData)
        self.reactor.register('GetScreenSize', self.GetScreenSize)
        self.reactor.register('Screenshot', self.Screenshot)
        self.reactor.register('Click', self.Click)
        self.reactor.register('Swipe', self.Swipe)
        self.reactor.register('LongClick', self.LongClick)
        self.reactor.register('KeyEvent', self.KeyEvent)
        self.reactor.register('SetForeground', self.SetForeground)
        self.reactor.register('ConnectWindow', self.ConnectWindow)
        self.reactor.register('Scroll', self.Scroll)
        self.reactor.register('RClick', self.RClick)
        self.reactor.register('DoubleClick', self.DoubleClick)
        transport = TcpSocket()
        transport.bind(self.addr)
        self.rpc = StdRpcEndpointController(transport, self.reactor)
        if self.running is False:
            self.running = True
            self.rpc.serve_forever()

if __name__ == '__main__':
    pocosdk = PocoSDKWindows()
    # pocosdk.ConnectWindow({"title_re": u".+?画图$"})
    # pocosdk.Swipe(0.3, 0.3, 0.8, 0.8, duration=5)
    pocosdk.run()



# coding=utf-8
__author__ = 'lxn3032'


import warnings
from functools import wraps


def deprecated(message):
    def wrapper(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            warnings.warn("Deprecation Warning: " + message)
            return func(*args, **kwargs)
        return wrapped
    return wrapper



""" JSON-RPC wrappers for version 1.0 and 2.0.

Objects diring init operation try to choose JSON-RPC 2.0 and in case of error
JSON-RPC 1.0.
from_json methods could decide what format is it by presence of 'jsonrpc'
attribute.

"""
from .utils import JSONSerializable
from .jsonrpc1 import JSONRPC10Request
from .jsonrpc2 import JSONRPC20Request


class JSONRPCRequest(JSONSerializable):

    """ JSONRPC Request."""

    @classmethod
    def from_json(cls, json_str):
        data = cls.deserialize(json_str)
        if isinstance(data, dict) and "jsonrpc" not in data:
            return JSONRPC10Request.from_json(json_str)
        else:
            return JSONRPC20Request.from_json(json_str)



# -*- coding: utf-8 -*-

from poco.drivers.std import StdPoco
from poco.drivers.ue4.device import UE4EditorWindow
from airtest.core.api import connect_device, device as current_device

__all__ = ['UE4Poco', 'DEFAULT_PORT', 'DEFAULT_ADDR']
DEFAULT_PORT = 5001
DEFAULT_ADDR = ('localhost', DEFAULT_PORT)


class UE4Poco(StdPoco):

    def __init__(self, addr=DEFAULT_ADDR, ue4_editor=False, connect_default_device=True, device=None, **options):
        if 'action_interval' not in options:
            options['action_interval'] = 0.5

        if ue4_editor:
            dev = UE4EditorWindow()
        else:
            dev = device or current_device()

        if dev is None and connect_default_device and not current_device():
            dev = connect_device("Android:///")

        super(UE4Poco, self).__init__(addr[1], dev, ip=addr[0], **options)
